//*****************************************************************************
//*
//* FILE NAME:  RACFResourceController.java
//*
//* IBM Confidential
//*
//*  OCO Source Materials
//*
//*  5647-A01
//*
//*  (C) Copyright IBM Corp. 2000
//*
//*   The source code for this program is not published or otherwise
//*   divested of its trade secrets, irrespective of what has been
//*   deposited with the U.S. Copyright Office.
//*
//*   Status = JMSI619
//*
//*****************************************************************************
//*
//* COMPONENT:  Guenter Dockweiler
//*
//* CLASSES:  RACFResourceController (externally documented: no)
//*
//* CHANGE HISTORY:
//*
//*   Change Date     UserID   Description
//*   -----  -------- -------- ------------------------------------------------
//*   c0001  07/01/99 DOCK     initial version
//*   c0100  03/08/00 schawer  changes for simulator
//*   c0101  05/31/00 schawer  changes for simulator
//*   c0102  08/28/00 schawer  remove refresh
//*   c0103  11/06/00 schawer  generate commit requiest only if no simulate
//*   c0104  11/06/00 schawer  adapt to new policies
//*   c0105  11/09/00 schawer  implement backup/restore
//*   c0106  11/21/00 schawer  implement instantiate/undoInstantiate
//*   c0200  08/07/01 schawer  change requested service to parallel sysplex
//*   c0201  08/24/01 schawer  disable generation of child requests (temporary)
//*   c0202  08/29/01 schawer  generate child request with same requested service name than request
//*   c0401  01/25/02 schawer  RACF no commit!
//*****************************************************************************
package com.ibm.irr.bind;

import java.util.*;

import com.ibm.cim.common.*;
import com.ibm.cim.util.*;
import com.ibm.cim.bindscheduler.*;
import com.ibm.cim.bind.*;
import com.ibm.cim.bind.schema.*;
import com.ibm.cim.directory.*;
import com.ibm.cim.host.*;
import com.ibm.irr.nls.*;
// import com.ibm.iea.bind.parmlib.*; -ws [c0101]

/** RACF Support Resource Controller
*/
/** @author G. Dockweiler */

public class RACFResourceController extends CimResourceController
{
  private static String ivCopyright = RACFCopyright.IBM_2000;
  private static String CLASS       = "RACFResourceController";
  private static int cvNumReq; // [c0200]

  // important instance data inherited from CimResourceController:
  // CimHistoryManager         ivHistoryManager
  // (set by CimRequestController in setHistoryManager() method)
  // CimDirectoryAccess        ivLDAPAccess
  // CimSchema                 ivSchema
  // CimSupporterObject        ivSupporterObject

  private RACFCommandBuilder   ivCommandBuilder;
  private RACFResourceAccess   ivResourceAccess;
  private RACFCommandOutput    ivCommandOutput;
  private RACFBindRequestParms ivRequestParms;
  private RACFPolicies         ivPolicies;
  private RACFPrimeData        ivRefreshData;
  private boolean              ivRefreshRequired;

  private CimMessageHandler    ivMessageHandler;

  private boolean              ivUserIsAuthorized;
  private boolean              ivRACFActive;
  private String               ivUserID;

  private boolean              ivObjectExists;
  private boolean              ivGenericExists;
  private String               ivGenericProfile;
  private boolean              ivAccessedObjectExists;
  private boolean              ivAccessedGenericExists;
  private String               ivAccessedGenericProfile;
  private boolean              ivClassExists;
  private boolean              ivRefUserExists;
  private boolean              ivRefGroupExists;

// [c0106 -ws]
/** This method is called during simulation of perform instantiate.
  * <dd><dl>
  *  <dt> <b>Usage:</b>
	*   This method is reserved for use by the msys framework.
  *  </dt>
  * </dl></dd>
  * @return CimSupportCompletion code
  * @exception CimException occurs if an exception is thrown by a nested method.*/
  public int simulateInstantiate() throws CimException {
	// --------------------------------------------
	// trace
	// --------------------------------------------
	if(CimExtendedTrace.cvTrcLvl) {
        cvTrace.information("simuateInstantiate() is faked in simulation!", "RACFResourceController", "simulateInstantiate" );
	}

	return CimSupportCompletionCode.OK;
  }

/** This method is called during simulation of undo instantiate.
  * <dd><dl>
  *  <dt> <b>Usage:</b>
	*   This method is reserved for use by the msys framework.
  *  </dt>
  * </dl></dd>
  * @return CimSupportCompletion code
  * @exception CimException occurs if an exception is thrown by a nested method.*/
  public int simulateUndoInstantiate() throws CimException {
	// --------------------------------------------
	// trace
	// --------------------------------------------
	if(CimExtendedTrace.cvTrcLvl) {
        cvTrace.information("simulateUndoInstantiate() is faked in simulation!", "RACFResourceController", "simulateUndoInstantiate" );
	}

	return CimSupportCompletionCode.OK;
  }

/** This method is called during perform instantiate.
  * <dd><dl>
  *  <dt> <b>Usage:</b>
	*   This method is reserved for use by the msys framework.
  *  </dt>
  * </dl></dd>
  * @return CimSupportCompletion code
  * @exception CimException occurs if an exception is thrown by a nested method.*/
  public int instantiate() throws CimException {
   	// --------------------------------------------
   	// trace
   	// --------------------------------------------
   	if(CimExtendedTrace.cvTrcLvl) {
        cvTrace.information("instantiate() is faked in simulation!", "RACFResourceController", "instantiate" );
   	}

   	return CimSupportCompletionCode.OK;
  }
/** This method is called during undo instantiate.
  * <dd><dl>
  *  <dt> <b>Usage:</b>
	*   This method is reserved for use by the msys framework.
  *  </dt>
  * </dl></dd>
  * @return CimSupportCompletion code
  * @exception CimException occurs if an exception is thrown by a nested method.*/
  public int undoInstantiate() throws CimException {
	// --------------------------------------------
	// trace
	// --------------------------------------------
	if(CimExtendedTrace.cvTrcLvl) {
        cvTrace.information("undoInstantiate() is faked in simulation!", "RACFResourceController", "undoInstantiate" );
	}

	return CimSupportCompletionCode.OK;
  }


// [c0105] -ws
 /** Set a backup of the resource.
   * To save a resource configuration level, the resource controller
   * can set a backup, using this method.
   * The supporter must be able, to restore this saved configuration
   * level, if the 'restoreBackup' method is called.
   * If a resource controller creates a backup, this action
   * overwrites a previously set backup.
   * This function may be optionally implemented.
   * @return CimSupportCompletion code
   * @exception CimException if an unrecouverable error occurs.*/
   public int setBackup() throws CimException {
   	// --------------------------------------------
   	// trace
   	// --------------------------------------------
   	if(CimExtendedTrace.cvTrcLvl) {
   	   cvTrace.information("setBackup() is faked in simulation!", "RACFResourceController", "setBackup" );
   	}

     return CimSupportCompletionCode.OK;
   	//return CimSupportCompletionCode.NOT_IMPLEMENTED;
   }

 /** Restore the previous set resource backup.
   * The resource controller use this method to restore a previously
   * saved resource configuration level.
   * Such a resource configuration level can be created by using the
   * 'setBackup' method.
   * The supporter must be able, to restore this saved configuration
   * level, if the 'restoreBackup' method is called.
   * The resource controller can only restore one previous resource
   * configuration level.
   * This function may be optionally implemented.
   * @return CimSupportCompletion code
   * @exception CimException if an unrecouverable error occurs.*/
   public int restoreBackup() throws CimException {
   	// --------------------------------------------
   	// trace
   	// --------------------------------------------
   	if(CimExtendedTrace.cvTrcLvl) {
   	   cvTrace.information("restoreBackup is faked in simulation!", "RACFResourceController", "restoreBackup" );
   	}

     return CimSupportCompletionCode.OK;
   	//return CimSupportCompletionCode.NOT_IMPLEMENTED;
   }




  // [c0100] -ws

 /** Check if a commit request can be executed.
   * This function may be optionally implemented.
   * @param request - original request from the Bind Scheduler.
   * The Managed Object of this request is contained in
   * this object.
   * @return CimSupportCompletion code
   * @exception CimException if an unrecouverable error occurs.*/
   public int checkCommitUpdate(CimRequest request) throws CimException {
     // --------------------------------------------
     // trace
     // --------------------------------------------
     if(CimExtendedTrace.cvTrcLvl) {
       cvTrace.information("RACFResourceController.checkCommitUpdate faked!");
     }

     request.setReturnInfo( CimFunction.CHECK_COMMIT,
                             0, 0, 0, "" );
     return CimSupportCompletionCode.OK;
     //return CimSupportCompletionCode.NOT_IMPLEMENTED;
   }

 /** Simulates a commit request.
   * This function may be optionally implemented.
   * @param request - original request from the Bind Scheduler.
   * The Managed Object of this request is contained in
   * this object.
   * @return CimSupportCompletion code
   * @exception CimException if an unrecouverable error occurs.*/
   public int simulateCommitUpdate(CimRequest request) throws CimException {
     // --------------------------------------------
     // trace
     // --------------------------------------------
     if(CimExtendedTrace.cvTrcLvl) {
       cvTrace.information("RACFResourceController.simulateCommitUpdate faked!");
     }

     request.setReturnInfo( CimFunction.SIMULATE_COMMIT,
                             0, 0, 0, "" );
     return CimSupportCompletionCode.OK;
     //return CimSupportCompletionCode.NOT_IMPLEMENTED;
   }

 /** Executes a commit request.
   * This function may be optionally implemented.
   * @param request - original request from the Bind Scheduler.
   * The Managed Object of this request is contained in
   * this object.
   * @return CimSupportCompletion code
   * @exception CimException if an unrecouverable error occurs.*/
   public int commitUpdate(CimRequest request) throws CimException {

     // --------------------------------------------
     // trace
     // --------------------------------------------
     if(CimExtendedTrace.cvTrcLvl) {
       cvTrace.information("RACFResourceController.commitUpdate faked!");
     }

     request.setReturnInfo( CimFunction.COMMIT,
                             0, 0, 0, "" );
     return CimSupportCompletionCode.OK;
     //return CimSupportCompletionCode.NOT_IMPLEMENTED;
   }







  /**Instantiate RACF Bind support specific object setup
   * @exception CimException
  */
  public void init()
  throws CimException
  {
    String METHOD = "init()";
    System.err.println( "RACFResourceController.init() entered" );
    // -------------------------------------------------------------------
    // ivLDAPAccess, ivSchema and ivSupporterObject are already set during
    // CimResourceController instantiation in initializeAttributes() method
    // -------------------------------------------------------------------
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,true,CLASS,METHOD); }

  try {

    // -------------------------------------------------------------------
    // Create my own objects
    // -------------------------------------------------------------------
    ivUserID           = CimJava.getUserId().toUpperCase();

    ivMessageHandler   = new CimMessageHandler(RACFConstants.MSG_CLASS,
                                               RACFConstants.MSG_PATH,
                                               ivSupporterObject.getNls());
    ivCommandBuilder   = new RACFCommandBuilder();
    ivResourceAccess   = new RACFResourceAccess(ivUserID);
    ivCommandOutput    = new RACFCommandOutput();
    ivRequestParms     = new RACFBindRequestParms();
    System.err.println( "RACFResourceController.init() ivLDAPAccess="+ivLDAPAccess );
    ivPolicies         = new RACFPolicies(ivLDAPAccess,
                                          ivSupporterObject.getPersistentConfiguration());
    ivRefreshData      = new RACFPrimeData(ivLDAPAccess,
                                           ivSupporterObject.getPersistentConfiguration(),
                                           ivCommandBuilder,
                                           ivResourceAccess,
                                           ivCommandOutput);
    ivRefreshRequired  = false;
    // -------------------------------------------------------------------
    // Determine whether RACF is active and
    //                   RACF subsystem is defined and started correctly
    // -------------------------------------------------------------------
    String  command         = ivCommandBuilder.checkSecurityManager();
    boolean commandResult   = ivResourceAccess.issueCommand(command,ivCommandOutput);
    ivRACFActive            = ivCommandOutput.isRACFActive();

    // -------------------------------------------------------------------
    // Check whether RACF subsystem is usable (required for command I/F)
    // -------------------------------------------------------------------
    if (ivCommandOutput.isRACFSubsystemActive() == false)
    {
      String[] fillIns = new String[2];
      fillIns[0] = RACFConstants.SECURITY_SERVICE;
      fillIns[1] = String.valueOf(ivCommandOutput.getReasonCode());
      throw new CimException(ivMessageHandler.getCimMessage(RACFMessages.SUBSYS_NOTACTIVE,fillIns));
    }

    // -------------------------------------------------------------------
    // Check whether active user is authorized to execute RACF commands
    // -------------------------------------------------------------------
    ivUserIsAuthorized = false;
    if (ivRACFActive)
    {
      ivUserIsAuthorized = this.checkUserAuthorization(ivUserID);
    }

  } catch (RACFInternalError e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    CimMessage message;
    try {
      message = ivMessageHandler.getCimMessage(RACFMessages.RACF_EXCEPTION,null);
    }
    catch (CimException emsg) {
      message = emsg.getCimMessage();
    }
    throw new CimException(message);
  }
  catch (RACFInconsistentOutput e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    CimMessage message;
    try {
      message = ivMessageHandler.getCimMessage(RACFMessages.RACF_EXCEPTION,null);
    }
    catch (CimException emsg) {
      message = emsg.getCimMessage();
    }
    throw new CimException(message);
  }
  catch (UnsatisfiedLinkError u) { // native DLL in RACFResourceAccess could not be loaded
    String[] fillIns = new String[2];
    fillIns[0] = RACFConstants.SECURITY_DLL;
    fillIns[1] = RACFConstants.SECURITY_DLLDIR;
    CimMessage message;
    try {
      message = ivMessageHandler.getCimMessage(RACFMessages.LOAD_DLL,fillIns);
    }
    catch (CimException emsg) {
      message = emsg.getCimMessage();
    }
    throw new CimException(message);
  }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,false,CLASS,METHOD); }
  }


  /** Get instance of extended trace object
   * @return the extended trace object for the actual job (refresh, prepareUpdate, performUpdate, etc)
  */
  static CimExtendedTrace getExtendedTrace()
  {
    return cvTrace;
  }


  /**Retrieve ID of security database to be used
   * @return shared security data base ID
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  // Security data base ID is specified as priming data
  // It enables the customer to distinguish between multiple security data bases,
  // e.g. when running update requests against multiple target systems with
  // different data bases.
  // -------------------------------------------------------------------
  public String getSharedResourceID(CimManagedObject managedObject)
  throws CimException
  {
    String METHOD = "getSharedResourceID()";
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,true,CLASS,METHOD); }

    String dataBaseID = "RACF.SHARED.DB";  // [c0104]

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Shared resource ID: "+dataBaseID,false,CLASS,METHOD); }
    return dataBaseID;
  }


  /**Retrieve actual value for given managed object
   * @param managedObject is the managed object for an update reuest
   * @return value folder record containing vaue folder basic for 'exists' entry
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  // Called before check/performUpdate/undoUpdate is called for the same request
  // (only if there are history relevant data for request)
  // Value in returned value folder reflects existence/non-existence of object
  // -------------------------------------------------------------------
  public CimValueFolder[] getActualValue
   (CimManagedObject managedObject)
  throws CimException
  {
    String               name;
    CimValueFolder[]     valueFolderArray;
    CimValueFolderRecord valueFolder;
    CimValueFolderBasic  existsValue;
    String               METHOD = "getActualValue()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Managed object ID: "+managedObject.getString(),true,CLASS,METHOD); }

  try {

    // -------------------------------------------------------------------
    // get ValueFolder record for global name
    // -------------------------------------------------------------------
    name             = managedObject.getGlobalName();
    valueFolderArray = new CimValueFolder[1];
    valueFolder      = (CimValueFolderRecord)ivSchema.createValueFolder(name);

    // -------------------------------------------------------------------
    // Only for requests other than SPECIAL_OPERATION we have an 'exists'
    // value. For SPECIAL_OPERATION return the ValueFolder retrieved above
    // untouched. The history controller also never will search an 'exists'
    // value there, because it is not defined in the Schema.
    // -------------------------------------------------------------------
    if (false == name.equals(RACFConstants.SUPP_SPECIALOP))
    {
      // -------------------------------------------------------------------
      // get ValueFolder to store actual value
      // -------------------------------------------------------------------
      existsValue      = (CimValueFolderBasic)valueFolder.getElementByName(RACFConstants.KEYWORD_EXISTS);

      // -------------------------------------------------------------------
      // check whether requested object already exists (-> ivObjectExists is set)
      // -------------------------------------------------------------------
      RACFBindRequestParms requestParms = new RACFBindRequestParms(managedObject);
      ivObjectExists = false;  // not authorized -> no existence check -> assume 'false'
      if (ivUserIsAuthorized)  // this is also the case if RACF is not active
      { this.checkObjectExistence(requestParms); }

      // -------------------------------------------------------------------
      // set value depending on check result
      // -------------------------------------------------------------------
      if (true == ivObjectExists)
      { existsValue.setValue("true"); }
      else
      { existsValue.setValue("false"); }

    } // end if other than SPECIAL_OPERATION

    // -------------------------------------------------------------------
    // return the value folder
    // -------------------------------------------------------------------
    valueFolderArray[0] = (CimValueFolder)valueFolder;

  } catch (RACFInternalError e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    valueFolderArray = null;
    CimMessage message;
    try {
      message = ivMessageHandler.getCimMessage(RACFMessages.RACF_EXCEPTION,null);
    }
    catch (CimException emsg) {
      message = emsg.getCimMessage();
    }
    throw new CimException(message);
  }
  catch (RACFInconsistentOutput e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    valueFolderArray = null;
    CimMessage message;
    try {
      message = ivMessageHandler.getCimMessage(RACFMessages.RACF_EXCEPTION,null);
    }
    catch (CimException emsg) {
      message = emsg.getCimMessage();
    }
    throw new CimException(message);
  }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Value folder",false,CLASS,METHOD); }
    return valueFolderArray;
  }

  /**Set priming data (RACF active/not active, shared RACF data baseID)
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  public int refresh()
  throws CimException
  {
    int     completionCode = CimSupportCompletionCode.OK;
    int     returnCode     = RACFConstants.RC_OK;
    int     reasonCode     = RACFConstants.RS_OK;
    String  shortMessage   = SHORT_MSG_OK;
    String  METHOD         = "refresh()";

    System.err.println( "RACFResourceController.refresh() entered" );
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,true,CLASS,METHOD); }

    // -------------------------------------------------------------------
    // Save RACF active indicator as priming data
    // -------------------------------------------------------------------
/* [c0102]
    ivRefreshData.setRACFActive(ivRACFActive);
  try {

    if (ivRACFActive)
    {
      // -------------------------------------------------------------------
      // Set RACF class data in public data (refresh) area
      // -------------------------------------------------------------------
      // ivRefreshData.setRACFClassData();  !!! must be activated for PS stage 2

      // -------------------------------------------------------------------
      // Make sure that the data set profile for the msys Parmlib datasets HLQ
      // is set up correctly
      // -------------------------------------------------------------------
      int result = assureParmlibHLQ();

      if (result == RACFConstants.RS_NOT_AUTHORIZED)
      {
        // -------------------------------------------------------------------
        // the user has no SPECIAL attribute
        // (it's required to performed commands in assureParmlibHLQ())
        // in this case write warning message into log file
        // -------------------------------------------------------------------
        reasonCode     = RACFConstants.RS_NOT_AUTHORIZED;
        shortMessage   = "SUCCESSFUL/NOT_AUTHORIZED";
        this.logMessage(RACFMessages.NOTAUTH_REFRESH,RACFConstants.SECURITY_SERVICE,
                        ivUserID,null,null,null);
      }
      if (result == RACFConstants.RS_POLICY_UPDATE)
      {
        // -------------------------------------------------------------------
        // 'no update' policy is set, i.e. user can not define/alter parmlib dataset profile
        // in this case write warning message into log file
        // -------------------------------------------------------------------
        reasonCode     = RACFConstants.RS_POLICY_UPDATE;
        shortMessage   = "SUCCESSFUL/NOUPDATE_POLICY";
        this.logMessage(RACFMessages.UPDPOL_NOREFRESH,RACFConstants.SECURITY_SERVICE,
                        null,null,null,null);
      }
    }

  } catch (RACFInternalError e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_ERROR;
    shortMessage   = "ERROR/RACF_ERROR";
  }
  catch (RACFInconsistentOutput e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_OUTPUT;
    shortMessage   = "ERROR/RACF_OUTPUT";
  }
  catch (CimException e) {
    completionCode = CimSupportCompletionCode.FAILED;
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_CIM_EXCEPTION;
    shortMessage   = "ERROR/CIM_EXCEPTION_CAUGHT";
  }
*/
    ivSupporterObject.setReturnInfo(completionCode,
                                    returnCode,
                                    reasonCode,
                                    shortMessage);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }

  /**Evaluate the update request: check syntactical correctness (if not
   *already done by RequestController using Schema) and check whether
   *request may be ignored or a child request must be generated
   * @param request is the actual update request
   * @return completion code
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  // We don't want to keep a history for requests that are not going
  // to be executed. This may even disturb the processing of the 'real'
  // requests.
  // -------------------------------------------------------------------
  public int prepareUpdate
   (CimRequest       request)
  throws CimException
  {
    int             completionCode     = CimSupportCompletionCode.OK;
    int             returnCode         = RACFConstants.RC_OK;
    int             reasonCode         = RACFConstants.RS_OK;
    String          shortMessage       = SHORT_MSG_OK;
    boolean         prepareComplete    = false;
    boolean         childGenerated     = false;

    String          oldRequestString   = "";
    String          newRequestString   = "";
    String          oldProfileName     = "";
    String          newProfileName     = "";
    String          METHOD             = "prepareUpdate()";

    System.err.println( "RACFResourceController.prepareUpdate() entered" );
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,true,CLASS,METHOD); }

  try {

    // -------------------------------------------------------------------
    // Get parameters for this request
    // -------------------------------------------------------------------
    ivRequestParms.actualizeRequestParms(request);

    // -------------------------------------------------------------------
    // Check request syntax
    // -------------------------------------------------------------------
    int syntaxError = ivRequestParms.checkSyntax();
    if (RACFConstants.RC_OK != syntaxError)
    {
      completionCode = CimSupportCompletionCode.FAILED;
      returnCode     = RACFConstants.RC_ERROR;
      reasonCode     = RACFConstants.RS_REQUEST_SYNTAX;
      shortMessage   = "ERROR/REQUEST_SYNTAX_ERROR";

      String messageID = RACFMessages.SYNTAX_PARM;
      if (RACFConstants.RS_REQUEST_SYNTAX_REQMODE == syntaxError)
      { messageID = RACFMessages.SYNTAX_REQPARM; }

      this.logMessage(messageID,ivRequestParms.toString(false),
                      ivRequestParms.valueInError(),ivRequestParms.keywordInError(),
                      null,request);

      prepareComplete = true;
    }

    // -------------------------------------------------------------------
    // 1. case to be ignored: RACF is not active
    // -------------------------------------------------------------------
    if (false == ivRACFActive     &&
        false == prepareComplete)
    {
      completionCode = CimSupportCompletionCode.NOP;
      returnCode     = RACFConstants.RC_IGNORED;
      reasonCode     = RACFConstants.RS_RACF_NOT_ACTIVE;
      shortMessage   = "IGNORED/RACF_NOT_ACTIVE";
      this.logMessage(RACFMessages.RACF_NOTACTIVE,ivRequestParms.toString(false),
                      buildDefineCommand(),null,null,request);
      prepareComplete = true;
    }
    // -------------------------------------------------------------------
    // 2. case to be ignored: execMode=IGNORE
    // -------------------------------------------------------------------
    if (ivRequestParms.execMode().equals(RACFConstants.MODE_IGNORE) &&
        false == prepareComplete)
    {
      completionCode = CimSupportCompletionCode.NOP;
      returnCode     = RACFConstants.RC_IGNORED;
      reasonCode     = RACFConstants.RS_IGNORE;
      shortMessage   = "IGNORED/EXECMODE_IGNORE";
      this.logMessage(RACFMessages.EXEC_IGNORE,ivRequestParms.toString(false),
                      null,null,null,request);
      prepareComplete = true;
    }

    // -------------------------------------------------------------------
    // if policy does not allow update, don't proceed any further
    // leave the request untouched and let performUpdate handle this case
    // -------------------------------------------------------------------
    if (false == ivPolicies.isUpdateAllowed())
    {
      prepareComplete = true;
    }

    // -------------------------------------------------------------------
    // check whether has SPECIAL attribute
    // it's required even if just LISTxxxx kind of command are to be performed,
    // e.g. to check for object existence
    // -------------------------------------------------------------------
    if (false == ivUserIsAuthorized &&
        false == prepareComplete)
    {
      completionCode = CimSupportCompletionCode.FAILED;
      returnCode     = RACFConstants.RC_ERROR;
      reasonCode     = RACFConstants.RS_NOT_AUTHORIZED;
      shortMessage   = "NO_EVALUATION/NOT_AUTHORIZED";
      this.logMessage(RACFMessages.NOTAUTH_PREPARE,ivRequestParms.toString(false),
                      RACFConstants.SECURITY_SERVICE,ivUserID,null,request);
      prepareComplete = true;
    }

    // -------------------------------------------------------------------
    // Actualize object existence parameters
    //  - ivObjectExists/ivAccessedObjectExists
    //  - ivGenericExists/ivAccessedGenericExists
    //  - ivGenericProfile/ivAccessedGenericProfile
    // -------------------------------------------------------------------
    if (false == prepareComplete)
    {
      if (false == ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_SPECIALOP))
      {
        oldRequestString = ivRequestParms.toString(true);
        oldProfileName   = ivRequestParms.resource();
        if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET))
        { oldProfileName = ivRequestParms.dataset(); }

        this.checkObjectExistence(ivRequestParms);
        this.finalizeRequestParms();

        newRequestString = ivRequestParms.toString(true);
        newProfileName   = ivRequestParms.resource();
        if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET))
        { newProfileName = ivRequestParms.dataset(); }
      }
    }

    // -------------------------------------------------------------------
    // If necessary, generate a child request to be processed instead of the
    // original one. This must happen if the request has been modified (e.g.
    // execMode or profile name changed). For conditions under which this happens
    // refer to method finalizeRequestParms().
    //
    // If the value of the 'resource'/'dataset' parameter is different for the
    // child request, a new managed object must be generated (the managed object
    // ID contains the changed value). Otherwise the managed object of the
    // actual request can be used.
    // -------------------------------------------------------------------
    if (false == oldRequestString.equals(newRequestString) &&
        false == prepareComplete)
    {
      // get correct managed object
      CimManagedObject managedObject;
      if (oldProfileName.equals(newProfileName))
      { managedObject = (CimManagedObject)request.getManagedObject(); }
      else
      { managedObject = ivRequestParms.newManagedObject(null); }

      // generate child request and add to actual request
      CimRequest childRequest = this.createNewRequest(ivRequestParms,request,managedObject,true); //[c0201]
      ivSupporterObject.addChildRequest(request,childRequest);                                    //[c0201]
      childGenerated = true;

      this.logMessage(RACFMessages.CHILD_REQUEST,oldRequestString,newRequestString,
                      null,null,request);
      prepareComplete = true;
    }

    // -------------------------------------------------------------------
    // Evaluation complete w/o generating a child request
    // -------------------------------------------------------------------
    if (false == childGenerated   &&
        false == prepareComplete)
    {
      this.logMessage(RACFMessages.PREPARE_OK,ivRequestParms.toString(true),
                      null,null,null,request);
    }

  } catch (RACFInternalError e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_ERROR;
    shortMessage   = "ERROR/RACF_ERROR";
  }
  catch (RACFInconsistentOutput e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_OUTPUT;
    shortMessage   = "ERROR/RACF_OUTPUT";
  }
  catch (CimException e) {
    completionCode = CimSupportCompletionCode.FAILED;
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_CIM_EXCEPTION;
    shortMessage   = "ERROR/CIM_EXCEPTION_CAUGHT";
  }

  try {
    request.setReturnInfo(ivSupporterObject.getFunction(),
                          completionCode,
                          returnCode,
                          reasonCode,
                          shortMessage);
  } catch (CimException e) {
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    this.logMessage(RACFMessages.NORETURN_INFO,Integer.toString(completionCode),Integer.toString(returnCode),
                    Integer.toString(reasonCode),shortMessage,request);
  }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }


  /**Check whether update request will run successfully
   * @param request is the actual update request
   * @param reUpdate is true if this is a re-do of the actual request
   * @param oldManagedObject is the managed object of the original update
   *request if it is a re-update for a different object, otherwise null
   * @return completion code
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  public int checkPerformUpdate
   (CimRequest       request,
    boolean          reUpdate,
    CimManagedObject oldManagedObject)
  throws CimException
  {
    int     completionCode         = CimSupportCompletionCode.OK;
    int     returnCode             = RACFConstants.RC_OK;
    int     reasonCode             = RACFConstants.RS_OK;
    String  shortMessage           = SHORT_MSG_OK;
    boolean objectExistsReferenced = false;
    boolean checkComplete          = false;
    int     numReferences;
    String  METHOD                 = "checkPerformUpdate()";
    System.err.println( "RACFResourceController.checkPerformUpdate() entered" );

    if (CimExtendedTrace.cvTrcLvl) {
      String info = "ReUpdate: "+reUpdate;
      if (oldManagedObject != null) {info = info+cvTrace.getIndentString()+
                                            "Old managed object ID: "+oldManagedObject.getString();}
      cvTrace.externalCall(info,true,CLASS,METHOD);
    }

  try {

    // -------------------------------------------------------------------
    // Get parameters for this request
    // -------------------------------------------------------------------
    ivRequestParms.actualizeRequestParms(request);

    // -------------------------------------------------------------------
    // If policy does not allow update, don't proceed any further
    // leave the request untouched and let performUpdate handle this case
    // -------------------------------------------------------------------
    if (false == ivPolicies.isUpdateAllowed())
    {
      checkComplete = true;
    }

    // -------------------------------------------------------------------
    // CHECK: user has not SPECIAL attribute
    //    -> request would fail
    // -------------------------------------------------------------------
    if (false == ivUserIsAuthorized &&
        false == checkComplete)
    {
      completionCode = CimSupportCompletionCode.WOULD_FAIL;
      returnCode     = RACFConstants.RC_CHECK;
      reasonCode     = RACFConstants.RS_NOT_AUTHORIZED;
      shortMessage   = "CHECK/NOT_AUTHORIZED";
      this.logMessage(RACFMessages.NOTAUTH_UPDATE,ivRequestParms.toString(false),
                      RACFConstants.SECURITY_SERVICE,ivUserID,null,request);
      checkComplete  = true;
    }

    // -------------------------------------------------------------------
    // For SPECIAL_OPERATION no further checking required
    // -------------------------------------------------------------------
    if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_SPECIALOP) &&
        false == checkComplete)
    {
      checkComplete  = true;
    }

    // -------------------------------------------------------------------
    // Actualize object existence parameters
    //  - ivObjectExists/ivAccessedObjectExists
    //  - ivGenericExists/ivAccessedGenericExists
    //  - ivGenericProfile/ivAccessedGenericProfile
    //  - ivRefUserExists/ivRefGroupExists
    // -------------------------------------------------------------------
    if (false == checkComplete)
    {
      this.checkObjectExistence(ivRequestParms);

      // -------------------------------------------------------------------
      // Check whether object is referenced already by other request
      // If so, handle object as existing
      //
      // At this point in time we also have to finally resolve the execMode
      // parameter value ASSURE2DEFINE. This value indicates that the original
      // execMode was ASSURE and the object did not exist during prepareUpdate().
      // Now we also know whether the object is referenced by a preceeding request
      // or whether it is defined in the (potentially big) time gap between
      // prepareUpdate() and now.
      //
      // If it is referenced by a preceeding request
      //   -> use execMode ALTER
      // otherwise
      //   if it exists now
      //     -> use execMode ALTER
      //   else
      //     -> use execMode DEFINE
      //
      // (the same will happen during performUpdate())
      // -------------------------------------------------------------------
      CimManagedObject managedObject = (CimManagedObject)request.getManagedObject();
      if (1 < ivHistoryManager.getValueRequestSetSize(managedObject))
      { // object is referenced by a preceeding request
        ivObjectExists         = true;
        objectExistsReferenced = true;

        if (ivRequestParms.execMode().equals(RACFConstants.MODE_ASSURE2DEFINE))
        {
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Finalizing execution mode "+
                                                               ivRequestParms.execMode()+
                                                               " with "+RACFConstants.MODE_ALTER+
                                                               " (object referenced by preceeding request)",CLASS,METHOD); }
          ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
        }
      }
      else
      { // object is NOT referenced by a preceeding request
        if (ivRequestParms.execMode().equals(RACFConstants.MODE_ASSURE2DEFINE))
        {
          if (true == ivObjectExists)
          { // object not referenced by a preceeding request, but created in the meantime
            if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Finalizing execution mode "+
                                                                 ivRequestParms.execMode()+
                                                                 " with "+RACFConstants.MODE_ALTER+
                                                                 " (object not referenced by preceeding request, but created in the meantime)",CLASS,METHOD); }
            ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
          }
          else
          { // object not referenced by a preceeding request and still not existing
            if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Finalizing execution mode "+
                                                                 ivRequestParms.execMode()+
                                                                 " with "+RACFConstants.MODE_DEFINE+
                                                                 " (object not referenced by preceeding request)",CLASS,METHOD); }
            ivRequestParms.overwriteExecMode(RACFConstants.MODE_DEFINE);
          }
        }
      }

      // -------------------------------------------------------------------
      // Is this a re-do of this request for the same object and that object
      // sill exists?
      //  -> even a DEFINE request must be handled as ALTER
      //     (the same will happen during performUpdate())
      // -------------------------------------------------------------------
      if (null == oldManagedObject &&
          true == reUpdate         &&
          true == ivObjectExists)
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+ivRequestParms.execMode()+
                                         " with "+RACFConstants.MODE_ALTER+" due to re-update operation",CLASS,METHOD); }
        ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
      }
    }

    if (false == checkComplete)
    {
    // -------------------------------------------------------------------
    // CHECK: specified class for ASSURE_RESOURCE/ACCESS does not exist
    //    -> request would fail
    // -------------------------------------------------------------------
      if ((ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_RESOURCE) ||
           ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS))     &&
          false == ivClassExists)
      {
        completionCode = CimSupportCompletionCode.WOULD_FAIL;
        returnCode     = RACFConstants.RC_CHECK;
        reasonCode     = RACFConstants.RS_CLASS_NOTEXISTS;
        shortMessage   = "CHECK/CLASS_NOTEXISTS";
        this.logMessage(RACFMessages.CLASS_NOTEXIST,ivRequestParms.toString(false),
                        ivRequestParms.rclass(),null,null,request);
        checkComplete  = true;
      }
    }

    if (false == checkComplete)
    {
    // -------------------------------------------------------------------
    // CHECK: user or group to be connected does not exist
    //    (not in RACF database and not referenced by preceeding update request)
    //    -> request would fail
    // -------------------------------------------------------------------
      if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_CONNECTION))
      {
        numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_USER));
        if (false == ivRefUserExists &&
            0     == numReferences)
        {
          completionCode = CimSupportCompletionCode.WOULD_FAIL;
          returnCode     = RACFConstants.RC_CHECK;
          reasonCode     = RACFConstants.RS_NO_USER_TO_CONNECT;
          shortMessage   = "CHECK/NO_USER_TO_CONNECT";
          this.logMessage(RACFMessages.USER_NOTEXIST,ivRequestParms.toString(false),
                          ivRequestParms.user(),null,null,request);
          checkComplete  = true;
        }
        numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_GROUP));
        if (false == ivRefGroupExists &&
            0     == numReferences)
        {
          completionCode = CimSupportCompletionCode.WOULD_FAIL;
          returnCode     = RACFConstants.RC_CHECK;
          reasonCode     = RACFConstants.RS_NO_GROUP_TO_CONNECT;
          shortMessage   = "CHECK/NO_GROUP_TO_CONNECT";
          this.logMessage(RACFMessages.GROUP_NOTEXIST,ivRequestParms.toString(false),
                          ivRequestParms.group(),null,null,request);
          checkComplete  = true;
        }
      }
    }

    if (false == checkComplete)
    {
    // -------------------------------------------------------------------
    // CHECK: user/group to which access permission is to be granted does not exist
    //    (not in RACF database and not referenced by preceeding update request)
    //    -> request would fail
    // -------------------------------------------------------------------
      if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS))
      {
        numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_USRGRP_USER));
        if (0 == numReferences)
        { numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_USRGRP_GROUP)); }
        if (false == ivRefUserExists  &&
            false == ivRefGroupExists &&
            0     == numReferences)
        {
          completionCode = CimSupportCompletionCode.WOULD_FAIL;
          returnCode     = RACFConstants.RC_CHECK;
          reasonCode     = RACFConstants.RS_NO_ACCESS_USRGRP;
          shortMessage   = "CHECK/USERORGROUP_NOTEXISTS";
          this.logMessage(RACFMessages.USRGRP_NOTEXIST,ivRequestParms.toString(false),
                          ivRequestParms.userOrGroup(),null,null,request);
          checkComplete  = true;
        }
      }
    }

    if (false == checkComplete)
    {
    // -------------------------------------------------------------------
    // CHECK: referenced profile for access permission does not exist
    //    (not in RACF database and not referenced by preceeding update request)
    //    and this is not tolerated (profileNotExists=FAIL)
    //    -> request would fail
    // -------------------------------------------------------------------
      if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS) &&
          ivRequestParms.resourceNotExists().equals(RACFConstants.NOTEXISTS_FAIL))
      {
        numReferences = 0;
        if (ivRequestParms.rclass().equals(RACFConstants.OBJ_DATASET))
        { numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_DATASET)); }
        else
        { numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_RESOURCE)); }

        if (false == ivAccessedObjectExists &&
            0     == numReferences)
        {
          completionCode = CimSupportCompletionCode.WOULD_FAIL;
          returnCode     = RACFConstants.RC_CHECK;
          reasonCode     = RACFConstants.RS_REF_RESOURCE;
          shortMessage   = "CHECK/REF_PROFILE_NOTEXISTS";
          this.logMessage(RACFMessages.NOREF_NODEF,ivRequestParms.toString(false),
                          ivRequestParms.resource(),null,null,request);
          checkComplete  = true;
        }
      }
    }

    // -------------------------------------------------------------------
    // Definition of a new user (or group) is also prohibited if a group (or user) with the
    // same name already exists
    // -------------------------------------------------------------------
    if (false == checkComplete)
    {
      if (ivRequestParms.execMode().equals(RACFConstants.MODE_DEFINE))
      {
        if ((ivRequestParms.objectType().equals(RACFConstants.SUPP_USER)  &&
             true == ivRefGroupExists)                                          ||
            (ivRequestParms.objectType().equals(RACFConstants.SUPP_GROUP) &&
             true == ivRefUserExists))
        { ivObjectExists = true; }
      }

      // -------------------------------------------------------------------
      // CHECK: execution mode DEFINE for existing object
      //    (existing may also mean that object is referenced by one of the previous requests)
      //    -> request would fail
      // -------------------------------------------------------------------
      if (ivRequestParms.execMode().equals(RACFConstants.MODE_DEFINE) &&
          true == ivObjectExists)
      {
        completionCode = CimSupportCompletionCode.WOULD_FAIL;
        returnCode     = RACFConstants.RC_CHECK;
        if (true == objectExistsReferenced)
        {
          reasonCode     = RACFConstants.RS_DEFINE_EXIST_REQUEST;
          shortMessage   = "CHECK/DEFINE_EXISTING_REF_OBJECT";
          this.logMessage(RACFMessages.OTHERREQ_NODEF,ivRequestParms.toString(false),
                          null,null,null,request);
        } else
        {
          reasonCode     = RACFConstants.RS_DEFINE_EXIST;
          shortMessage   = "CHECK/DEFINE_EXISTING_OBJECT";
          this.logMessage(RACFMessages.EXIST_NODEF,ivRequestParms.toString(false),
                          null,null,null,request);
        }
        checkComplete  = true;
      }

    // -------------------------------------------------------------------
    // CHECK: execution mode ALTER for not existing object
    //    (not existing also means that object is not referenced by one of the previous requests)
    //    -> request would fail
    // -------------------------------------------------------------------
      if (ivRequestParms.execMode().equals(RACFConstants.MODE_ALTER) &&
          false == ivObjectExists)
      {
        completionCode = CimSupportCompletionCode.WOULD_FAIL;
        returnCode     = RACFConstants.RC_CHECK;
        reasonCode     = RACFConstants.RS_ALTER_NOTEXIST;
        shortMessage   = "CHECK/ALTER_NOTEXISTING_OBJECT";
        this.logMessage(RACFMessages.NOTEXIST_NOALT,ivRequestParms.toString(false),
                        null,null,null,request);
        checkComplete  = true;
      }

    }

  } catch (RACFInternalError e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_ERROR;
    shortMessage   = "ERROR/RACF_ERROR";
  }
  catch (RACFInconsistentOutput e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_OUTPUT;
    shortMessage   = "ERROR/RACF_OUTPUT";
  }
  catch (CimException e) {
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_CIM_EXCEPTION;
    shortMessage   = "ERROR/CIM_EXCEPTION_CAUGHT";
  }

  try {
    request.setReturnInfo(ivSupporterObject.getFunction(),
                          completionCode,
                          returnCode,
                          reasonCode,
                          shortMessage);
  } catch (CimException e) {
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    this.logMessage(RACFMessages.NORETURN_INFO,Integer.toString(completionCode),Integer.toString(returnCode),
                    Integer.toString(reasonCode),shortMessage,request);
  }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }


  /**Update processing which is common for performUpdate() and simulatePerformUpdate()
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  private int commonUpdate
   (CimRequest       request,
    boolean          reUpdate,
    CimManagedObject oldManagedObject,
    boolean          simulate)
  throws CimException
  {
    int     completionCode    = CimSupportCompletionCode.OK;
    int     returnCode        = RACFConstants.RC_OK;
    int     reasonCode        = RACFConstants.RS_OK;
    String  shortMessage      = SHORT_MSG_OK;
    boolean noDeleteExecution = false;
    boolean noUpdateExecution = false;
    boolean commandResult;
    boolean updateComplete    = false;
    String  command           = "";
    String  METHOD            = "commonUpdate()";

    System.err.println( "RACFResourceController.commonUpdate() entered" );

    if (CimExtendedTrace.cvTrcLvl) {
      String info = "Reupdate: "+reUpdate;
      if (oldManagedObject != null) {info = info+cvTrace.getIndentString()+
                                            "Old managed object ID: "+oldManagedObject.getString();}
      info = info + cvTrace.getIndentString()+ "Simulate: "+simulate;
      cvTrace.externalCall(info,true,CLASS,METHOD);
    }

  try {

    // -------------------------------------------------------------------
    // If it's a re-do of this request for a different object,
    // the old object must be deleted (only done if policy allows this)
    // -------------------------------------------------------------------
    if (null != oldManagedObject            &&
        true == reUpdate                    &&
        true == ivPolicies.isUpdateAllowed())
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.controlFlow("Entered path: Reupdate for different object",
                                       "Old object will be deleted (undone) if it still exists"); }
      // reconstruct old request parameters from managed object
      // and check old object existence
      RACFBindRequestParms oldRequestParms = new RACFBindRequestParms(oldManagedObject);
      this.checkObjectExistence(oldRequestParms);

      // the old object was defined by old request and still exists
      // -> build the corresponding DELETE request and re-use undoUpdate() logic
      if (true == ivObjectExists)
      {
        CimRequest deleteRequest = this.createNewRequest(oldRequestParms,request,oldManagedObject,false);
        completionCode = this.commonUndoUpdate(deleteRequest,simulate,true);

        // if the undo failed, we have to stop here and declare this update request as failed
        if (CimSupportCompletionCode.FAILED == completionCode)
        {
          returnCode     = RACFConstants.RC_ERROR;
          reasonCode     = RACFConstants.RS_REUPDATE_UNDO;
          shortMessage   = "ERROR/UNDO_DURING_RE-UPDATE_FAILED";

          if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
          return completionCode;
        }
      }

    }

    // -------------------------------------------------------------------
    // Get parameters for this request
    // -------------------------------------------------------------------
    ivRequestParms.actualizeRequestParms(request);

    // -------------------------------------------------------------------
    // Policiy does not allow execution of this request
    // -------------------------------------------------------------------
    if (false == ivPolicies.isUpdateAllowed())
    {
      returnCode     = RACFConstants.RC_NO_EXECUTION;
      reasonCode     = RACFConstants.RS_POLICY_UPDATE;
      shortMessage   = "NO_EXECUTION/NOUPDATE_POLICY";
      this.logMessage(RACFMessages.UPDPOL_NOUPDATE,ivRequestParms.toString(false),
                      RACFConstants.SECURITY_SERVICE,null,null,request);
      updateComplete = true;
    }

    if (false == updateComplete)
    {
      // -------------------------------------------------------------------
      // Actualize object existence parameters
      //  - ivObjectExists/ivAccessedObjectExists
      //  - ivGenericExists/ivAccessedGenericExists
      //  - ivGenericProfile/ivAccessedGenericProfile
      // -------------------------------------------------------------------
      this.checkObjectExistence(ivRequestParms);

      // -------------------------------------------------------------------
      // If the referenced profile for access permission does not exist
      // (not in RACF database and not referenced by preceeding update request)
      // and this is tolerated (profileNotExists=ACCEPT)
      // -> do not process this request
      // -------------------------------------------------------------------
      if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS) &&
          ivRequestParms.resourceNotExists().equals(RACFConstants.NOTEXISTS_ACCEPT))
      {
        int numReferences = 0;
        if (ivRequestParms.rclass().equals(RACFConstants.OBJ_DATASET))
        { numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_DATASET)); }
        else
        { numReferences = ivHistoryManager.getValueRequestSetSize(ivRequestParms.newManagedObject(RACFConstants.OBJ_RESOURCE)); }

        if (false == ivAccessedObjectExists &&
            0     == numReferences)
        {
          reasonCode     = RACFConstants.RS_REF_RESOURCE;
          shortMessage   = "IGNORED/PROFILE_NOTEXISTS_ACCEPTED";
          this.logMessage(RACFMessages.NOREF_IGNORE,ivRequestParms.toString(false),
                          ivRequestParms.resource(),null,null,request);
          updateComplete = true;
        }
      }
    }

    if (false == updateComplete)
    {
      // -------------------------------------------------------------------
      // At this point in time we also have to finally resolve the execMode
      // parameter value ASSURE2DEFINE. This value indicates that the original
      // execMode was ASSURE and the object did not exist during prepareUpdate().
      // Now we know whether the object has been created by a preceeding request
      // and we also know whether the object is defined in the (potentially big)
      // time gap between prepareUpdate() and now.
      //
      // If it is referenced by a preceeding request
      //   -> use execMode ALTER
      // otherwise
      //   if it exists now
      //     -> use execMode ALTER
      //   else
      //     -> use execMode DEFINE
      // -------------------------------------------------------------------
      CimManagedObject managedObject = (CimManagedObject)request.getManagedObject();
      if (1 < ivHistoryManager.getValueRequestSetSize(managedObject))
      { // object is referenced by a preceeding request
        ivObjectExists         = true;

        if (ivRequestParms.execMode().equals(RACFConstants.MODE_ASSURE2DEFINE))
        {
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Finalizing execution mode "+
                                                               ivRequestParms.execMode()+
                                                               " with "+RACFConstants.MODE_ALTER+
                                                               " (object referenced by preceeding request)",CLASS,METHOD); }
          ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
        }
      }
      else
      { // object is NOT referenced by a preceeding request
        if (ivRequestParms.execMode().equals(RACFConstants.MODE_ASSURE2DEFINE))
        {
          if (true == ivObjectExists)
          { // object not referenced by a preceeding request, but created in the meantime
            if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Finalizing execution mode "+
                                                                 ivRequestParms.execMode()+
                                                                 " with "+RACFConstants.MODE_ALTER+
                                                                 " (object not referenced by preceeding request, but created in the meantime)",CLASS,METHOD); }
            ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
          }
          else
          { // object not referenced by a preceeding request and still not existing
            if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Finalizing execution mode "+
                                                                 ivRequestParms.execMode()+
                                                                 " with "+RACFConstants.MODE_DEFINE+
                                                                 " (object not referenced by preceeding request)",CLASS,METHOD); }
            ivRequestParms.overwriteExecMode(RACFConstants.MODE_DEFINE);
          }
        }
      }

      // -------------------------------------------------------------------
      // Is this a re-do of this request for the same object and that object
      // sill exists?
      //  -> even a DEFINE request must be handled as ALTER
      // -------------------------------------------------------------------
      if (null == oldManagedObject &&
          true == reUpdate         &&
          true == ivObjectExists)
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+
                                                             ivRequestParms.execMode()+
                                                             " with "+RACFConstants.MODE_ALTER+
                                                             " due to re-update operation on existing object",CLASS,METHOD); }
        ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
      }

      // -------------------------------------------------------------------
      // Build the actual command to be executed (during prepareUpdate() it is
      // assured that execMode is always either ALTER or DEFINE)
      // -------------------------------------------------------------------
      if (ivRequestParms.execMode().equals(RACFConstants.MODE_DEFINE))
      { command = buildDefineCommand(); }
      else
      { command = buildAlterCommand(); }

    }

    // -------------------------------------------------------------------
    // in simulate mode just log the command
    // -------------------------------------------------------------------
    if (false == updateComplete &&
        true  == simulate)
    {
      this.logMessage(RACFMessages.SIMULATE_UPDATE,command,
                      ivRequestParms.toString(false),null,null,request);
    }

    // -------------------------------------------------------------------
    // in non-simulate mode execute the command
    // -------------------------------------------------------------------
    if (false == updateComplete &&
        false == simulate)
    {
      // any error situation that's not already covered by one of the checks
      // above is unexpected
      if (false == ivResourceAccess.issueCommand(command,ivCommandOutput))
      {
        completionCode = CimSupportCompletionCode.FAILED;
        returnCode     = RACFConstants.RC_ERROR;
        reasonCode     = RACFConstants.RS_COMMAND_ERROR;
        shortMessage   = "ERROR/COMMAND_ERROR";
        this.logMessage(RACFMessages.CMDERR_NOUPDATE,command,
                        ivRequestParms.toString(false),null,null,request);
        this.logMessage(RACFMessages.COMMAND_RCS,Integer.toString(ivCommandOutput.getResultNative()),
                        Integer.toString(ivCommandOutput.getSAFRC()),Integer.toString(ivCommandOutput.getRACFRC()),
                        Integer.toString(ivCommandOutput.getReasonCode()),request);
        String[] messages = ivCommandOutput.getRACFErrorMessages();
        this.logMessage(RACFMessages.FAIL_MSGS,messages[0],messages[1],messages[2],null,request);
      }
      else
      {
        this.logMessage(RACFMessages.UPDATE_CMDOK,command,
                        ivRequestParms.toString(false),null,null,request);
      }

    }

  } catch (RACFInternalError e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_ERROR;
    shortMessage   = "ERROR/RACF_ERROR";
  }
  catch (RACFInconsistentOutput e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_OUTPUT;
    shortMessage   = "ERROR/RACF_OUTPUT";
  }
  catch (CimException e) {
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_CIM_EXCEPTION;
    shortMessage   = "ERROR/CIM_EXCEPTION_CAUGHT";
  }

  try {
    request.setReturnInfo(ivSupporterObject.getFunction(),
                          completionCode,
                          returnCode,
                          reasonCode,
                          shortMessage);
  } catch (CimException e) {
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    this.logMessage(RACFMessages.NORETURN_INFO,Integer.toString(completionCode),Integer.toString(returnCode),
                    Integer.toString(reasonCode),shortMessage,request);
  }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }





     //-- add new request
     if( !simulate &&
         !request.getRequestedServiceName().equals( "RACFSecurity" ) ) { // [c0401]
        // [c0100]
        cvNumReq++;
        CimRequest newRequest = new CimRequest(
                                     CimFunction.COMMIT,
                                     "DEMO_COMMIT_BIND",
                                     //0==(cvNumReq%2) ? "ParallelSysplex" : "RACFSecurity", //  [c0200]
                                     request.getRequestedServiceName(),     // [c0202]
                                     "COMMIT",
                                     RACFConstants.SECURITY_VERSION,
                                     null,
                                     request );
        ivSupporterObject.addChildRequest(request, newRequest);
        System.err.println( "RACFResourceController.commonUpdate() created request="+newRequest );
     }








    return completionCode;
  }


  /**Execute an update requests
   * @param request is the actual update request
   * @param reUpdate is true if this is a re-do of the actual request
   * @param oldManagedObject is the managed object of the original update
   *request if it is a re-update for a different object, otherwise null
   * @return completion code
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  public int performUpdate
   (CimRequest       request,
    boolean          reUpdate,
    CimManagedObject oldManagedObject)
  throws CimException
  {
    String  METHOD = "performUpdate()";

    if (CimExtendedTrace.cvTrcLvl) {
      String info = "Reupdate: "+reUpdate;
      if (oldManagedObject != null) {info = info+cvTrace.getIndentString()+
                                            "Old managed object ID: "+oldManagedObject.getString();}
      cvTrace.externalCall(info,true,CLASS,METHOD);
    }

    int completionCode = this.commonUpdate(request,reUpdate,oldManagedObject,false);

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }


  /**Simulate an update requests
   * @param request is the actual update request
   * @param reUpdate is true if this is a re-do of the actual request
   * @param oldManagedObject is the managed object of the original update
   *request if it is a re-update for a different object, otherwise null
   * @return completion code
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  public int simulatePerformUpdate
   (CimRequest       request,
    boolean          reUpdate,
    CimManagedObject oldManagedObject)
  throws CimException
  {
    String  METHOD = "simulatePerformUpdate()";

    if (CimExtendedTrace.cvTrcLvl) {
      String info = "Reupdate: "+reUpdate;
      if (oldManagedObject != null) {info = info+cvTrace.getIndentString()+
                                            "Old managed object ID: "+oldManagedObject.getString();}
      cvTrace.externalCall(info,true,CLASS,METHOD);
    }

    int completionCode = this.commonUpdate(request,reUpdate,oldManagedObject,true);

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }


  /**Check whether undo update request will run successfully
   * @param request is the actual update request
   * @return completion code
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  public int checkUndoUpdate
   (CimRequest request)
  throws CimException
  {
    boolean checkComplete     = false;
    int     completionCode    = CimSupportCompletionCode.OK;
    int     returnCode        = RACFConstants.RC_OK;
    int     reasonCode        = RACFConstants.RS_OK;
    String  shortMessage      = SHORT_MSG_OK;
    String  METHOD            = "checkUndoUpdate()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,true,CLASS,METHOD); }

    // -------------------------------------------------------------------
    // If policy does not allow update or delete, don't proceed any further
    // let undoUpdate handle this case
    // -------------------------------------------------------------------
    if (false == ivPolicies.isUpdateAllowed()  ||
        false == ivPolicies.isDeletionAllowed())
    {
      checkComplete = true;
    }

    // -------------------------------------------------------------------
    // CHECK: user has not SPECIAL attribute
    //    -> undo request would fail
    // -------------------------------------------------------------------
    if (false == ivUserIsAuthorized &&
        false == checkComplete)
    {
      completionCode = CimSupportCompletionCode.WOULD_FAIL;
      returnCode     = RACFConstants.RC_CHECK;
      reasonCode     = RACFConstants.RS_NOT_AUTHORIZED;
      shortMessage   = "CHECK/NOT_AUTHORIZED";
      this.logMessage(RACFMessages.NOTAUTH_UNDO,ivRequestParms.toString(false),
                      RACFConstants.SECURITY_SERVICE,ivUserID,null,request);
    }

  try {
    request.setReturnInfo(ivSupporterObject.getFunction(),
                          completionCode,
                          returnCode,
                          reasonCode,
                          shortMessage);
  } catch (CimException e) {
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    this.logMessage(RACFMessages.NORETURN_INFO,Integer.toString(completionCode),Integer.toString(returnCode),
                    Integer.toString(reasonCode),shortMessage,request);
  }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }


  /**Und update processing which is common for undoUpdate() and simulateUndoUpdate()
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  private int commonUndoUpdate
   (CimRequest request,
    boolean    simulate,
    boolean    redoUndo)
  throws CimException
  {
    String  METHOD = "commonUndoUpdate()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Simulate: "+simulate,true,CLASS,METHOD); }

    int              completionCode    = CimSupportCompletionCode.OK;
    int              returnCode        = RACFConstants.RC_OK;
    int              reasonCode        = RACFConstants.RS_OK;
    String           shortMessage      = SHORT_MSG_OK;
    boolean          undoComplete      = false;
    String           command;
    CimManagedObject managedObject     = (CimManagedObject)request.getManagedObject();

  try {

    // -------------------------------------------------------------------
    // Get parameters for this request
    // -------------------------------------------------------------------
    ivRequestParms.actualizeRequestParms(request);

    // -------------------------------------------------------------------
    // Check whether object was created by this request ...
    // -------------------------------------------------------------------
    if (false == undoComplete)
    {

      String value;
      CimValueFolderRecord valueFolder = (CimValueFolderRecord)ivHistoryManager.getValue(managedObject);
      if (valueFolder != null)
      {
        CimValueFolderBasic existsValue = (CimValueFolderBasic)valueFolder.getElementByName(RACFConstants.KEYWORD_EXISTS);
        value = existsValue.getValue();
        if (value == null)
        { value = "true"; }
      }
      else
      {
        this.logMessage(RACFMessages.NO_HISTORY,managedObject.toString(),
                        null,null,null,request);
        value = "true";
      }

      // -------------------------------------------------------------------
      // ... no, it existed already before => do not delete it
      // -------------------------------------------------------------------
      if (value.equals("true"))
      {
        returnCode     = RACFConstants.RC_NO_EXECUTION;
        reasonCode     = RACFConstants.RS_ALREADY_EXISTED;
        shortMessage   = "NO_EXECUTION/EXISTED_BEFORE_UPDATE";
        this.logMessage(RACFMessages.EXIST_NOUNDO,ivRequestParms.toString(false),
                        null,null,null,request);
        undoComplete = true;
      }

    }
    // -------------------------------------------------------------------
    // Finally we decided to really undo the request, i.e. object deletion
    // -------------------------------------------------------------------
    if (false == undoComplete)
    {

      command = this.buildDeleteCommand(ivRequestParms);

      // -------------------------------------------------------------------
      // in simulate mode just log the delete command corresponding to undo update
      // -------------------------------------------------------------------
      if (simulate)
      {
        this.logMessage(RACFMessages.SIMULATE_UNDO,command,
                        ivRequestParms.toString(false),null,null,request);
      }
      // -------------------------------------------------------------------
      // in non-simulate mode execute the command corresponding to undo update
      // -------------------------------------------------------------------
      else
      {
        // -------------------------------------------------------------------
        // before executing the delete command check whether policies allow it
        // -------------------------------------------------------------------
        if (false == ivPolicies.isDeletionAllowed())
        {
          returnCode     = RACFConstants.RC_NO_EXECUTION;
          reasonCode     = RACFConstants.RS_POLICY_DELETE;
          shortMessage   = "NO_EXECUTION/NODELETE_POLICY";
          this.logMessage(RACFMessages.DELPOL_NOUNDO,ivRequestParms.toString(false),
                          RACFConstants.SECURITY_SERVICE,null,null,request);
          undoComplete = true;
        }
        if (false == ivPolicies.isUpdateAllowed())
        {
          returnCode     = RACFConstants.RC_NO_EXECUTION;
          reasonCode     = RACFConstants.RS_POLICY_UPDATE;
          shortMessage   = "NO_EXECUTION/NOUPDATE_POLICY";
          this.logMessage(RACFMessages.UPDPOL_NOUNDO,ivRequestParms.toString(false),
                          RACFConstants.SECURITY_SERVICE,null,null,request);
          undoComplete = true;
        }

        // --------------------------------------------------------------------------
        // execute the delete command corresponding to undo update
        // --------------------------------------------------------------------------
        if (false == undoComplete)
        {
          if (false == ivResourceAccess.issueCommand(command,ivCommandOutput))
          {
            if (true == ivCommandOutput.deletedObjectDoesNotExist
                        (ivRequestParms.objectIndicator(ivRequestParms.objectType())))
            {
              reasonCode     = RACFConstants.RS_ALREADY_DELETED;
              shortMessage   = "SUCCESSFUL/OBJECT_ALREADY_DELETED";
              this.logMessage(RACFMessages.GONE_NOUNDO,ivRequestParms.toString(false),
                              null,null,null,request);
            }
            else
            {
              completionCode = CimSupportCompletionCode.FAILED;
              returnCode     = RACFConstants.RC_ERROR;
              reasonCode     = RACFConstants.RS_COMMAND_ERROR;
              shortMessage   = "ERROR/COMMAND_ERROR";
              this.logMessage(RACFMessages.CMDERR_NOUNDO,command,
                              ivRequestParms.toString(false),null,null,request);
              this.logMessage(RACFMessages.COMMAND_RCS,Integer.toString(ivCommandOutput.getResultNative()),
                              Integer.toString(ivCommandOutput.getSAFRC()),Integer.toString(ivCommandOutput.getRACFRC()),
                              Integer.toString(ivCommandOutput.getReasonCode()),request);
              String[] messages = ivCommandOutput.getRACFErrorMessages();
              this.logMessage(RACFMessages.FAIL_MSGS,messages[0],messages[1],messages[2],null,request);
            }
          }
          else  // command exectued successfully
          {
            this.logMessage(RACFMessages.UNDO_CMDOK,command,
                            ivRequestParms.toString(false),null,null,request);
          }

        } // no execution because of policy

      } // end simulate check

    } // end execution/no execution check

  } catch (RACFInternalError e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_ERROR;
    shortMessage   = "ERROR/RACF_ERROR";
  }
  catch (RACFInconsistentOutput e)
  {
    e.logTheError(ivMessageHandler,ivSupporterObject);
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_RACF_OUTPUT;
    shortMessage   = "ERROR/RACF_OUTPUT";
  }
  catch (CimException e) {
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
    ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
    completionCode = CimSupportCompletionCode.FAILED;
    returnCode     = RACFConstants.RC_ERROR;
    reasonCode     = RACFConstants.RS_CIM_EXCEPTION;
    shortMessage   = "ERROR/CIM_EXCEPTION_CAUGHT";
  }

    // if we are called just for undo update in case of re-update with different oject ID,
    // the request is only generated internally (does not exist in LDAP)
    // => we can not set the return info, just pass completion code back to update logic
    if (false == redoUndo)
    {
      try {
        request.setReturnInfo(ivSupporterObject.getFunction(),
                              completionCode,
                              returnCode,
                              reasonCode,
                              shortMessage);
      } catch (CimException e) {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.traceException(e); }
        ivSupporterObject.getSupporterLog().logAdditionalInfo(e.getMessage());
        this.logMessage(RACFMessages.NORETURN_INFO,Integer.toString(completionCode),Integer.toString(returnCode),
                        Integer.toString(reasonCode),shortMessage,request);
      }
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }




     //-- add new request
     if( !simulate &&
         !request.getRequestedServiceName().equals( "RACFSecurity" ) ) { // [c0401]
        // [c0100]
        cvNumReq++;
        CimRequest newRequest = new CimRequest(
                                     CimFunction.COMMIT,
                                     "DEMO_COMMIT_UNBIND",
                                     //0==(cvNumReq%2) ? "ParallelSysplex" : "RACFSecurity", //  [c0200]
                                     request.getRequestedServiceName(),  // [c0202]
                                     "COMMIT",
                                     RACFConstants.SECURITY_VERSION,
                                     null,
                                     request );
        ivSupporterObject.addChildRequest(request, newRequest);
        System.err.println( "RACFResourceController.commonUndoUpdate() created request="+newRequest );
     }




    return completionCode;
  }


  /**Undo an update request
   * @param request is the actual update request
   * @return completion code
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  // Undo the update request if
  //  - security policies allow object deletion
  //  - object was originally created by the request
  // -------------------------------------------------------------------
  public int undoUpdate
   (CimRequest request)
  throws CimException
  {
    String  METHOD = "undoUpdate()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,true,CLASS,METHOD); }

    int completionCode = this.commonUndoUpdate(request,false,false);

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }


  /**Simulate undoing an update request
   * @param request is the actual update request
   * @return completion code
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  // Simulate undoing an update request
  // -------------------------------------------------------------------
  public int simulateUndoUpdate
   (CimRequest       request)
  throws CimException
  {
    String  METHOD = "simlateUndoUpdate()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall(null,true,CLASS,METHOD); }

    int completionCode = this.commonUndoUpdate(request,true,false);

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.externalCall("Completion code: "+completionCode,false,CLASS,METHOD); }
    return completionCode;
  }


  /**Build the command to check whether an object exists
  */
  // -------------------------------------------------------------------
  // requestParms has been created from managedObjectID of object to be checked
  // -------------------------------------------------------------------
  private String buildCheckExistenceCommand
   (RACFBindRequestParms requestParms)
  {
    String  METHOD = "buildCheckExistenceCommand()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,true,CLASS,METHOD); }

    String command = "";

    if (requestParms.objectType().equals(RACFConstants.SUPP_USER))
    {
      command = ivCommandBuilder.checkUserExistence(requestParms.user());
    }
    else if (requestParms.objectType().equals(RACFConstants.SUPP_GROUP))
    {
      command = ivCommandBuilder.checkGroupExistence(requestParms.group());
    }
    else if (requestParms.objectType().equals(RACFConstants.SUPP_RESOURCE))
    {
      command = ivCommandBuilder.checkResourceExistence(requestParms.rclass(),
                                                        requestParms.resource());
    }
    else if (requestParms.objectType().equals(RACFConstants.SUPP_DATASET))
    {
      command = ivCommandBuilder.checkDatasetExistence(requestParms.dataset());
    }
    else if (requestParms.objectType().equals(RACFConstants.SUPP_ACCESS))
    {
      command = ivCommandBuilder.checkAccessExistence(requestParms.rclass(),
                                                      requestParms.resource());
    }
    else if (requestParms.objectType().equals(RACFConstants.SUPP_CONNECTION))
    { // check user existence command also used to check connection existence
      command = ivCommandBuilder.checkUserExistence(requestParms.user());
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Command: "+command,false,CLASS,METHOD); }
    return command;
  }


  /**Build the command to alter an object
  */
  // -------------------------------------------------------------------
  // determine object type from requested support name and object name
  // and parameters from actual request parameters
  // -------------------------------------------------------------------
  private String buildAlterCommand()
  {
    String  METHOD = "buildAlterCommand()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,true,CLASS,METHOD); }

    String command = "";

    if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_USER))
    {
      command = ivCommandBuilder.alterUser(ivRequestParms.user(),
                                           ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_GROUP))
    {
      command = ivCommandBuilder.alterGroup(ivRequestParms.group(),
                                            ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_RESOURCE))
    {
      command = ivCommandBuilder.alterResource(ivRequestParms.rclass(),
                                               ivRequestParms.resource(),
                                               ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET))
    {
      command = ivCommandBuilder.alterDataset(ivRequestParms.dataset(),
                                              ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS))
    {
      command = ivCommandBuilder.addAlterAccess(ivRequestParms.rclass(),
                                                ivRequestParms.resource(),
                                                ivRequestParms.userOrGroup(),
                                                ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_CONNECTION))
    {
      command = ivCommandBuilder.addAlterConnection(ivRequestParms.user(),
                                                    ivRequestParms.group(),
                                                    ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_SPECIALOP))
    {
      command = ivCommandBuilder.specialOperation(ivRequestParms.commandParameters());
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Command: "+command,false,CLASS,METHOD); }
    return command;
  }


  /**Build the command to define an object
  */
  // -------------------------------------------------------------------
  // determine object type from requested support name and object name
  // and parameters from actual request parameters
  // -------------------------------------------------------------------
  private String buildDefineCommand()
  {
    String  METHOD = "buildDefineCommand()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,true,CLASS,METHOD); }

    String command = "";

    if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_USER))
    {
      command = ivCommandBuilder.addUser(ivRequestParms.user(),
                                         ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_GROUP))
    {
      command = ivCommandBuilder.addGroup(ivRequestParms.group(),
                                          ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_RESOURCE))
    {
      command = ivCommandBuilder.addResource(ivRequestParms.rclass(),
                                             ivRequestParms.resource(),
                                             ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET))
    {
      command = ivCommandBuilder.addDataset(ivRequestParms.dataset(),
                                            ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS))
    {
      command = ivCommandBuilder.addAlterAccess(ivRequestParms.rclass(),
                                                ivRequestParms.resource(),
                                                ivRequestParms.userOrGroup(),
                                                ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_CONNECTION))
    {
      command = ivCommandBuilder.addAlterConnection(ivRequestParms.user(),
                                                    ivRequestParms.group(),
                                                    ivRequestParms.commandParameters());
    }

    else if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_SPECIALOP))
    {
      command = ivCommandBuilder.specialOperation(ivRequestParms.commandParameters());
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Command: "+command,false,CLASS,METHOD); }
    return command;
  }


  /**Build the command to delete objects in case of request re-execution for
    *different oject ID
  */
  // -------------------------------------------------------------------
  // requestParms has been created from managedObjectID of object to be deleted
  // -------------------------------------------------------------------
  private String buildDeleteCommand
   (RACFBindRequestParms requestParms)
  {
    String  METHOD = "buildDeleteCommand()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,true,CLASS,METHOD); }

    String command = "";

    if (requestParms.objectType().equals(RACFConstants.SUPP_USER))
    {
      command = ivCommandBuilder.deleteUser(requestParms.user());
    }

    else if (requestParms.objectType().equals(RACFConstants.SUPP_GROUP))
    {
      command = ivCommandBuilder.deleteGroup(requestParms.group());
    }

    else if (requestParms.objectType().equals(RACFConstants.SUPP_RESOURCE))
    {
      command = ivCommandBuilder.deleteResource(requestParms.rclass(),
                                                requestParms.resource());
    }

    else if (requestParms.objectType().equals(RACFConstants.SUPP_DATASET))
    {
      command = ivCommandBuilder.deleteDataset(requestParms.dataset());
    }

    else if (requestParms.objectType().equals(RACFConstants.SUPP_ACCESS))
    {
      command = ivCommandBuilder.deleteAccess(requestParms.rclass(),
                                              requestParms.resource(),
                                              requestParms.userOrGroup());
    }

    else if (requestParms.objectType().equals(RACFConstants.SUPP_CONNECTION))
    {
      command = ivCommandBuilder.deleteConnection(requestParms.user(),
                                                  requestParms.group());
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Command: "+command,false,CLASS,METHOD); }
    return command;
  }


  /**Transform a generic profile name into a discrete profile name
  */
  // -------------------------------------------------------------------
  // RES.NAME.HUGO  -> RES.NAME.HUGO (already discrete)
  // RES.NAME.HUGO* -> RES.NAME.HUGO
  // RES.NAME.*     -> RES.NAME
  // RES.NAME.**    -> RES.NAME
  // -------------------------------------------------------------------
  private String makeDiscreteProfile
   (String genericProfile)
  {
    String  METHOD = "makeDiscreteProfile()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Generic profile: "+genericProfile,true,CLASS,METHOD); }

    int[]  suffices = {'*', '*', '.'};
    String discreteProfile = genericProfile;
    int    index;

    for (int i=0; i<suffices.length; i++)
    {
      index = discreteProfile.lastIndexOf(suffices[i]);
      if (index == discreteProfile.length()-1)
      {
        discreteProfile = discreteProfile.substring(0,index);
      }
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Discrete profile: "+discreteProfile,false,CLASS,METHOD); }
    return discreteProfile;
  }


  /**Check whether requested object already exists
   * @exception RACFInconsistentOutput if the result of a RACF command does not match any of the expected output cases
   * @exception RACFInternalError if execution of a RACF command fails with an unexpected error
  */
  // -------------------------------------------------------------------
  //  For SPECIAL no check at all
  //  For USER check user existence
  //  For GROUP check group existence
  //  For RESOURCE check resource existence
  //      ...
  //
  //  For RESOURCE or DATASET also (if requested profile does not exist)
  //   - check existence of generic profile matching discrete profile
  //     (e.g. THIS.RES.NAME requested and THIS.RES.* exists)
  //   - save generic profile name if it exists
  //  This is needed to handle execMode=ALTER_GENERIC|ASSURE_GENERIC correctly
  //
  //  For ACCESS also
  //   - check whether the accessed profile (RESOURCE or DATASET) exists
  //   - if the accessed profile does not exist,
  //     check whether corresponding generic profile exists
  //   - save generic profile name if it exists
  //  This is needed to handle resourceNotExists=USE_GENERIC correctly
  // -------------------------------------------------------------------
  private void checkObjectExistence
   (RACFBindRequestParms requestParms)
   throws RACFInconsistentOutput, RACFInternalError
  {
    String           command;
    String           object;
    boolean          commandResult;
    String           METHOD = "checkObjectEistence()";

    ivObjectExists           = false;
    ivGenericExists          = false;
    ivGenericProfile         = null;
    ivAccessedObjectExists   = false;
    ivAccessedGenericExists  = false;
    ivAccessedGenericProfile = null;
    ivClassExists            = false;
    ivRefUserExists          = false;
    ivRefGroupExists         = false;

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,true,CLASS,METHOD); }

    //  For SPECIAL no check at all
    if (true == requestParms.requestedSupport().equals(RACFConstants.SUPP_SPECIALOP))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,false,CLASS,METHOD); }
      return;
    }

    command = buildCheckExistenceCommand(requestParms);

    commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);

    // inspect command output for USER existence
    if (requestParms.objectType().equals(RACFConstants.SUPP_USER))
    {
      ivObjectExists = ivCommandOutput.userExists(requestParms.user());

      // a GROUP with the same name as the USER must also not exist (checked in checkPerformUpdate())
      command          = ivCommandBuilder.checkGroupExistence(requestParms.user());
      commandResult    = ivResourceAccess.issueCommand(command,ivCommandOutput);
      ivRefGroupExists = ivCommandOutput.groupExists(requestParms.user());
    }

    // inspect command output for GROUP existence
    else if (requestParms.objectType().equals(RACFConstants.SUPP_GROUP))
    {
      ivObjectExists = ivCommandOutput.groupExists(requestParms.group());

      // a USER with the same name as the GROUP must also not exist (checked in checkPerformUpdate())
      command          = ivCommandBuilder.checkUserExistence(requestParms.group());
      commandResult    = ivResourceAccess.issueCommand(command,ivCommandOutput);
      ivRefGroupExists = ivCommandOutput.userExists(requestParms.group());
    }

    // inspect command output for RESOURCE existence
    else if (requestParms.objectType().equals(RACFConstants.SUPP_RESOURCE))
    {
      ivObjectExists = ivCommandOutput.resourceExists(requestParms.rclass(),
                                                      requestParms.resource());
      ivClassExists  = ivCommandOutput.classExists();
    }

    // inspect command output for DATASET existence
    else if (requestParms.objectType().equals(RACFConstants.SUPP_DATASET))
    {
      ivObjectExists = ivCommandOutput.datasetExists(requestParms.dataset());
    }

    // inspect command output for ACCESS existence
    // check also whether the user or group to be permitted exist
    else if (requestParms.objectType().equals(RACFConstants.SUPP_ACCESS))
    {
      ivObjectExists = ivCommandOutput.accessExists(requestParms.userOrGroup());
      if (requestParms.rclass().equals(RACFConstants.OBJ_DATASET))
      { ivClassExists  = true; }
      else
      { ivClassExists  = ivCommandOutput.classExists(); }

      command         = ivCommandBuilder.checkUserExistence(requestParms.userOrGroup());
      commandResult   = ivResourceAccess.issueCommand(command,ivCommandOutput);
      ivRefUserExists = ivCommandOutput.userExists(requestParms.userOrGroup());

      command          = ivCommandBuilder.checkGroupExistence(requestParms.userOrGroup());
      commandResult    = ivResourceAccess.issueCommand(command,ivCommandOutput);
      ivRefGroupExists = ivCommandOutput.groupExists(requestParms.userOrGroup());

      // for ACCESS, it is also of relevance whether or not the accessed profile exists
      // accessed profile may be discrete or generic
      if (requestParms.rclass().equals(RACFConstants.OBJ_DATASET))
      {
        command = ivCommandBuilder.checkDatasetExistence(requestParms.resource());
        commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);
        ivAccessedObjectExists = ivCommandOutput.datasetExists(requestParms.resource());
        if (false == ivAccessedObjectExists)
        {
          command = ivCommandBuilder.checkGenericDatasetExistence(this.makeDiscreteProfile(requestParms.resource()));
          commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);
          ivAccessedGenericProfile = ivCommandOutput.getGenericDataset();
        }
      }
      else
      {
        command = ivCommandBuilder.checkResourceExistence(requestParms.rclass(),
                                                          requestParms.resource());
        commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);
        ivAccessedObjectExists = ivCommandOutput.resourceExists(requestParms.rclass(),
                                                                requestParms.resource());
        if (false == ivAccessedObjectExists)
        {
          command = ivCommandBuilder.checkGenericResourceExistence(requestParms.rclass(),
                                                                   this.makeDiscreteProfile(requestParms.resource()));
          commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);
          ivAccessedGenericProfile = ivCommandOutput.getGenericResource(requestParms.rclass());
        }
      }
      if (null != ivAccessedGenericProfile)
      {
        ivAccessedGenericExists = true;
      }

    }

    // inspect command output for CONNECTION existence
    // check also whether the user and the group to be connected exist
    else if (requestParms.objectType().equals(RACFConstants.SUPP_CONNECTION))
    {
      ivObjectExists = ivCommandOutput.connectionExists(requestParms.group());

      command         = ivCommandBuilder.checkUserExistence(requestParms.user());
      commandResult   = ivResourceAccess.issueCommand(command,ivCommandOutput);
      ivRefUserExists = ivCommandOutput.userExists(requestParms.user());

      command          = ivCommandBuilder.checkGroupExistence(requestParms.group());
      commandResult    = ivResourceAccess.issueCommand(command,ivCommandOutput);
      ivRefGroupExists = ivCommandOutput.groupExists(requestParms.group());

    } // anything else is an unknown request or request for which no values are stored

    // if the object does not exist, for RESOURCE or DATASET there may be a generic profile
    if (false == ivObjectExists)
    {
      // check for generic RESOURCE existence
      if (requestParms.objectType().equals(RACFConstants.SUPP_RESOURCE))
      {
        command = ivCommandBuilder.checkGenericResourceExistence(requestParms.rclass(),
                                                                 this.makeDiscreteProfile(requestParms.resource()));
        commandResult    = ivResourceAccess.issueCommand(command,ivCommandOutput);
        ivGenericProfile = ivCommandOutput.getGenericResource(requestParms.rclass());
      }
      // check for generic DATASET existence
      else if (requestParms.objectType().equals(RACFConstants.SUPP_DATASET))
      {
        command = ivCommandBuilder.checkGenericDatasetExistence(this.makeDiscreteProfile(requestParms.dataset()));
        commandResult    = ivResourceAccess.issueCommand(command,ivCommandOutput);
        ivGenericProfile = ivCommandOutput.getGenericDataset();
      }
      if (null != ivGenericProfile)
      {
        ivGenericExists = true;
      }
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.information(cvTrace.getIndentString()+
                                     "Object exists:                   "+ivObjectExists+cvTrace.getIndentString()+
                                     "Generic profile exists:          "+ivGenericExists+cvTrace.getIndentString()+
                                     "Generic profile:                 "+ivGenericProfile+cvTrace.getIndentString()+
                                     "Accessed profile exists:         "+ivAccessedObjectExists+cvTrace.getIndentString()+
                                     "Accessed generic profile exists: "+ivAccessedGenericExists+cvTrace.getIndentString()+
                                     "Accessed generic profile:        "+ivAccessedGenericProfile+cvTrace.getIndentString()+
                                     "Class exists:                    "+ivClassExists+cvTrace.getIndentString()+
                                     "Referenced user exists:          "+ivRefUserExists+cvTrace.getIndentString()+
                                     "Referenced group exists:         "+ivRefGroupExists,CLASS,METHOD); }
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,false,CLASS,METHOD); }
  }


  /**Check whether active user is authorized to execute RACF commands
   * @exception RACFInconsistentOutput if the result of a RACF command does not match any of the expected output cases
   * @exception RACFInternalError if execution of a RACF command fails with an unexpected error
  */
  // -------------------------------------------------------------------
  private boolean checkUserAuthorization
   (String userID)
   throws RACFInconsistentOutput, RACFInternalError
  {
    String  METHOD = "checkUserAuthorization()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("user ID: "+userID,true,CLASS,METHOD); }

    String  command       = ivCommandBuilder.checkAuthorization(userID);
    boolean commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);
    boolean check         = ivCommandOutput.userIsAuthorized();

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Check result: "+check,false,CLASS,METHOD); }
    return check;
  }


  /**Transform execution modes ALTER_GENERIC, ASSURE_GENERIC and ASSURE
    *into execution mode ALTER or DEFINE and replace not-existing discrete
    *profile with generic profile
  */
  // -------------------------------------------------------------------
  // NOTE: If execMode is set tentatively to DEFINE (ASSURE2DEFINE), the
  //       final decision whether to use DEFINE or ALTER is made in
  //       checkPerformUpdate() and performUpdate
  //       Now we only know whether the object exists in the RACF database, during
  //       checkPerformUpdate() and performUpdate we also know whether it is referenced
  //       by a preceeding update request (because we hav a history available
  //       at that point). Being referenced by a preceding request is another
  //       indication for using ALTER instead of DEFINE.
  private void finalizeRequestParms()
  {
    String METHOD = "finalizeRequestParms()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,true,CLASS,METHOD); }

    // For execution mode ALTER_GENERIC:
    //  - Set execution mode to ALTER
    //  - If requested profile does not exist AND generic profile exists
    //    -> Use generic instead of requested profile (resource or dataset)
    if (ivRequestParms.execMode().equals(RACFConstants.MODE_ALTER_GENERIC))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+ivRequestParms.execMode()+
                                       " with "+RACFConstants.MODE_ALTER,CLASS,METHOD); }
      ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
      if (false == ivObjectExists &&
          true  == ivGenericExists)
      {
        if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_RESOURCE))
        {
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting resource profile name "+ivRequestParms.resource()+
                                           " with generic profile name "+ivGenericProfile,CLASS,METHOD); }
          ivRequestParms.overwriteResource(ivGenericProfile);
        }
        if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET))
        {
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting dataset profile name "+ivRequestParms.dataset()+
                                           " with generic profile name "+ivGenericProfile,CLASS,METHOD); }
          ivRequestParms.overwriteDataset(ivGenericProfile);
        }
      }
    }

    // For execution mode ASSURE_GENERIC:
    //  - If requested profile does not exist AND generic profile exists
    //    -> Set execution mode to ALTER
    //    -> Use generic instead of requested profile (resource or dataset)
    //  - If requested profile does not exist AND generic profile does not exist
    //    -> Set execution mode to DEFINE (tentative)
    //  - If requested profile exists
    //    -> Set execution mode to ALTER
    if (ivRequestParms.execMode().equals(RACFConstants.MODE_ASSURE_GENERIC))
    {
      if (false == ivObjectExists &&
          true  == ivGenericExists)
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+ivRequestParms.execMode()+
                                         " with "+RACFConstants.MODE_ALTER,CLASS,METHOD); }
        ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);

        if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_RESOURCE))
        {
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting resource profile name "+ivRequestParms.resource()+
                                           " with generic profile name "+ivGenericProfile,CLASS,METHOD); }
          ivRequestParms.overwriteResource(ivGenericProfile);
        }
        if (ivRequestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET))
        {
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting dataset profile name "+ivRequestParms.dataset()+
                                           " with generic profile name "+ivGenericProfile,CLASS,METHOD); }
          ivRequestParms.overwriteDataset(ivGenericProfile);
        }
      }
      if (false == ivObjectExists &&
          false == ivGenericExists)
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+ivRequestParms.execMode()+
                                         " with "+RACFConstants.MODE_ASSURE2DEFINE+" (finally resolved during perform update)",CLASS,METHOD); }
        ivRequestParms.overwriteExecMode(RACFConstants.MODE_ASSURE2DEFINE);
      }
      if (true == ivObjectExists)
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+ivRequestParms.execMode()+
                                         " with "+RACFConstants.MODE_ALTER,CLASS,METHOD); }
        ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
      }
    }

    // For execution mode ASSURE:
    //  - If object exists
    //    -> Set execution mode to ALTER
    //  - If object does not exist
    //    -> Set execution mode to DEFINE (tentative)
    if (ivRequestParms.execMode().equals(RACFConstants.MODE_ASSURE))
    {
      if (true == ivObjectExists)
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+ivRequestParms.execMode()+
                                         " with "+RACFConstants.MODE_ALTER,CLASS,METHOD); }
        ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
      }
      else
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting execution mode "+ivRequestParms.execMode()+
                                         " with "+RACFConstants.MODE_ASSURE2DEFINE+" (finally resolved during perform update)",CLASS,METHOD); }
        ivRequestParms.overwriteExecMode(RACFConstants.MODE_ASSURE2DEFINE);
      }
    }

    // For resourceNotExists = USE_GENERIC:
    //  - Set resourceNotExists to FAIL
    //  - If accessed profile does not exist, but a corresponding generic
    //    profile to be used for the access does exist
    //    -> Use generic profile instead of requested profile
    //    -> If original execution mode was ASSURE
    //       -> Set execution mode to ALTER
    if (ivRequestParms.resourceNotExists().equals(RACFConstants.NOTEXISTS_USE_GENERIC))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting resource_not_exists "+RACFConstants.NOTEXISTS_USE_GENERIC+
                                       " with "+RACFConstants.NOTEXISTS_FAIL,CLASS,METHOD); }
      ivRequestParms.overwriteNotExists(RACFConstants.NOTEXISTS_FAIL);

      if(false == ivAccessedObjectExists   &&
         true  == ivAccessedGenericExists)
      {
        if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Overwriting name of accessed profile "+ivRequestParms.resource()+
                                         " with generic profile name "+ivAccessedGenericProfile,CLASS,METHOD); }
        ivRequestParms.overwriteResource(ivAccessedGenericProfile);

        if (ivRequestParms.execMode().equals(RACFConstants.MODE_ASSURE2DEFINE))
        {
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("RE-Overwriting execution mode "+ivRequestParms.execMode()+
                                           " with "+RACFConstants.MODE_ALTER,CLASS,METHOD); }
          ivRequestParms.overwriteExecMode(RACFConstants.MODE_ALTER);
        }
      }
    }

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,false,CLASS,METHOD); }
  }


  /**Create a CimRequest to be used to delete an object (re-update request
    *for different object that still exists)
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  private CimRequest createNewRequest
   (RACFBindRequestParms requestParms,
    CimRequest           request,
    CimManagedObject     managedObject,
    boolean              childRequest)
  throws CimException
  {
    String  METHOD = "createNewRequest()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Requested support: "+requestParms.requestedSupport()+
                                     cvTrace.getIndentString()+"Managed object: "+managedObject.getString(),
                                     true,CLASS,METHOD); }

    CimKeywordValue[] keywordValueArray;
    CimKeywordValue   keywordValue;
    CimRequest        newRequest;
    int               parmCount;

    parmCount = 0;
    // how many parameters do we need?
    if (requestParms.requestedSupport().equals(RACFConstants.SUPP_USER)    ||
        requestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET) ||
        requestParms.requestedSupport().equals(RACFConstants.SUPP_GROUP))
    {
      parmCount = 3;
    }
    else if (requestParms.requestedSupport().equals(RACFConstants.SUPP_RESOURCE)   ||
             requestParms.requestedSupport().equals(RACFConstants.SUPP_CONNECTION))
    {
      parmCount = 4;
    }
    else if (requestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS))
    {
      parmCount = 6;
    }
    keywordValueArray = new CimKeywordValue[parmCount];

    // fill up the keyword/value array
    parmCount = 0;
    // add 'user' parameter if required
    if (requestParms.requestedSupport().equals(RACFConstants.SUPP_USER)       ||
        requestParms.requestedSupport().equals(RACFConstants.SUPP_CONNECTION))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_USER+" = "+requestParms.user()); }
      keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_USER,
                                         requestParms.user());
      keywordValueArray[parmCount] = keywordValue;
      parmCount++;
    }

    // add 'group' parameter if required
    if (requestParms.requestedSupport().equals(RACFConstants.SUPP_GROUP)      ||
        requestParms.requestedSupport().equals(RACFConstants.SUPP_CONNECTION))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_GROUP+" = "+requestParms.group()); }
      keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_GROUP,
                                         requestParms.group());
      keywordValueArray[parmCount] = keywordValue;
      parmCount++;
    }

     // add 'resource/class' parameter if required
    if (requestParms.requestedSupport().equals(RACFConstants.SUPP_RESOURCE) ||
        requestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_RESOURCE+" = "+requestParms.resource()); }
      keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_RESOURCE,
                                         requestParms.resource());
      keywordValueArray[parmCount] = keywordValue;
      parmCount++;
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_CLASS+" = "+requestParms.rclass()); }
      keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_CLASS,
                                         requestParms.rclass());
      keywordValueArray[parmCount] = keywordValue;
      parmCount++;
    }

    // add 'dataset' parameter if required
    if (requestParms.requestedSupport().equals(RACFConstants.SUPP_DATASET))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_DATASET+" = "+requestParms.dataset()); }
      keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_DATASET,
                                         requestParms.dataset());
      keywordValueArray[parmCount] = keywordValue;
      parmCount++;
    }

    // add 'userOrGroup/resourceNotExists' parameter if required
    if (requestParms.requestedSupport().equals(RACFConstants.SUPP_ACCESS))
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_USERORGROUP+" = "+requestParms.userOrGroup()); }
      keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_USERORGROUP,
                                         requestParms.userOrGroup());
      keywordValueArray[parmCount] = keywordValue;
      parmCount++;
      String value = requestParms.resourceNotExists();
      if (value.equals(""))
      {
        value = RACFConstants.NOTEXISTS_ACCEPT;
      }
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_RESOURCENOTEXISTS+" = "+value); }
      keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_RESOURCENOTEXISTS,
                                         value);
      keywordValueArray[parmCount] = keywordValue;
      parmCount++;
    }

    // add 'commandParameters/execMode' parameter (for all requests)
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_COMMANDPARAMETERS+" = "+requestParms.commandParameters()); }
    keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_COMMANDPARAMETERS,
                                       requestParms.commandParameters());
    keywordValueArray[parmCount] = keywordValue;
    parmCount++;
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.verboseInfo("Adding keyword/value "+RACFConstants.KEYWORD_EXECMODE+" = "+requestParms.execMode()); }
    keywordValue = new CimKeywordValue(RACFConstants.KEYWORD_EXECMODE,
                                       requestParms.execMode());
    keywordValueArray[parmCount] = keywordValue;

    String ID = "IRRCHILD";
    if (false == childRequest)
    { ID = "IRRDELETE"; }

    cvNumReq++;
    newRequest = new CimRequest(ivSupporterObject.getFunction(),
                                ID,
                                //0==(cvNumReq%2) ? "ParallelSysplex" : "RACFSecurity", //  [c0200]
                                request.getRequestedServiceName(),     // [c0202]
                                requestParms.requestedSupport(),
                                RACFConstants.SECURITY_VERSION,
                                keywordValueArray,
                                request);
    newRequest.setManagedObject(managedObject);

    if (CimExtendedTrace.cvTrcLvl)
    {
      cvTrace.information("New update request "+ID+" generated",CLASS,METHOD);
      cvTrace.internalCall(null,false,CLASS,METHOD);
    }
    return newRequest;
  }


  /**Make sure that the data set profile for the msys Parmlib datasets HLQ
   * is set up correctly
   * @exception CimException thrown in case of unexpected errors
   * @exception RACFInconsistentOutput if the result of a RACF command does not match any of the expected output cases
   * @exception RACFInternalError if execution of a RACF command fails with an unexpected error
  */
  // -------------------------------------------------------------------
  // If this routine is called and the Parmlib datasets are already allocated,
  // nothing has to be done. Reason: RAFC suport is always called before
  // Parmlib support. If called for the very first time, the correct setup
  // must beassured. All subsequent times no action is required, because the
  // setup has been done on the first call, the Parmlib datasets have been
  // allocated on the first call of Parmlib support and changing the HLQ has
  // no effect (Parmlib support always uses the datasets allocated the first
  // time -> no change in dataset profile setup).
  //
  // The fact whether the Parmlib datasets are already allocated (-> nothing
  // to do) is determined from the Parmlib support's private LDAP area.
  //
  // If the user has not SPECIAL attribute (it's required to performed commands),
  // return (false)
  //
  // If the datasets are not yet allocated, the generic dataset profile for
  // specifid HLQ is checked. If it exists, it is assured that it has
  // UACC(READ). If it does not exist, it is created with UACC(READ).
  //
  // This makes sure that anyone (including started tasks) accessing the msys
  // parmlib has read access, e.g. during IPL.
  // -------------------------------------------------------------------
  private boolean assureParmlibHLQ()
  throws CimException, RACFInternalError, RACFInconsistentOutput
  {
    boolean result = true;
    String  PARMLIB_POLICY_NAME       = "msysParmlib";
    String  PARMLIB_POLICY_CATEGORY   = "IeaParmlib";
    String  HLQ_POLICY_KEYWORD        = "parmlibHLQ";
    String  METHOD                    = "assureParmlibHLQ()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,true,CLASS,METHOD); }

    // if parmlib DS already allocated -> no action
    // ask the parmlib support to get that information
/* [c0101]
    IeaParmlibLDAPAccess parmlibLDAP = new IeaParmlibLDAPAccess(ivSupporterObject);
    String parmlibDSName = parmlibLDAP.getSysplexProperty(IeaParmlibLDAPAccess.msysParmlibName);

    if (parmlibDSName.equals(""))  // not yet allocated
    {
      if (false == ivUserIsAuthorized)
      {
        result = false;
      }
      else
      {
        // determine actual value of 'parmlib HLQ name' from Parmlib policy
        CimPolicy policy = new CimPolicy(ivLDAPAccess,
                                         ivSupporterObject.getPersistentConfiguration(),
                                         PARMLIB_POLICY_NAME,
                                         PARMLIB_POLICY_CATEGORY);

        String parmlibHLQ = policy.getValueForKeyword(HLQ_POLICY_KEYWORD)
                          + ".MSYS.**";

        // check whether generic dataset profile for msys Parmlib HLQ exists
        String  command       = ivCommandBuilder.checkDatasetExistence(parmlibHLQ);
        boolean commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);

        // if profile exists, make sure it has UACC(READ), else define it with UACC(READ)
        if (ivCommandOutput.datasetExists(parmlibHLQ))
        {
          command = ivCommandBuilder.alterDataset(parmlibHLQ,"UACC(READ)");
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Parmlib generic profile already exists"+
                                           " -> alter to UACC(READ)",CLASS,METHOD); }
        } else
        {
          command = ivCommandBuilder.addDataset(parmlibHLQ,"UACC(READ)");
          if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Defining Parmlib generic profile",CLASS,METHOD); }
        }
        commandResult = ivResourceAccess.issueCommand(command,ivCommandOutput);
        if (commandResult)
        {
          this.logMessage(RACFMessages.PARMLIB_DSPROF,RACFConstants.SECURITY_SERVICE,
                          command,null,null,null);
        }
      }
    }
    else
    {
      if (CimExtendedTrace.cvTrcLvl) { cvTrace.information("Parmlib support datasets already allocated"+
                                       " -> no action required",CLASS,METHOD); }
    }
[c0101] */
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,false,CLASS,METHOD); }
    return result;
  }
  /**Write message to log file
   * @exception CimException thrown in case of unexpected errors
  */
  // -------------------------------------------------------------------
  private void logMessage
   (String     messageKey,
    String     fillIn1,
    String     fillIn2,
    String     fillIn3,
    String     fillIn4,
    CimRequest request)
  throws CimException
  {
    String  METHOD = "logMessage()";

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall("Message key: "+messageKey+cvTrace.getIndentString()+
                                    "Fill-in #1: "+fillIn1+cvTrace.getIndentString()+
                                    "Fill-in #2: "+fillIn2+cvTrace.getIndentString()+
                                    "Fill-in #3: "+fillIn3+cvTrace.getIndentString()+
                                    "Fill-in #4: "+fillIn4+cvTrace.getIndentString(),true,CLASS,METHOD); }

    CimMessage message;
    String[]   fillIns   = new String[CimMessageHandler.MAXNUM_FILLINS];
    String     requestDN = "n/a";

    // initialize fill-in's
    for (int i=0; i<CimMessageHandler.MAXNUM_FILLINS; i++)
    {
      fillIns[i] = null;
    }
    fillIns[0] = fillIn1;
    fillIns[1] = fillIn2;
    fillIns[2] = fillIn3;
    fillIns[3] = fillIn4;

    // retrieve the message
    try {
      message = ivMessageHandler.getCimMessage(messageKey,fillIns);
    }
    catch (CimException e) {
      message = e.getCimMessage();
    }

    // write message to Supporter Object or request log file
    if (null != request)
    { requestDN = request.getRequestDN(); }
    if (CimExtendedTrace.cvTrcLvl) { cvTrace.logFileMessage (requestDN,message.getKey(),
                                                                       message.getText(true),
                                                                       message.getExplanation(false),
                                                                       message.getSystemAction(false),
                                                                       message.getUserAction(false)); }
    ivSupporterObject.getSupporterLog().logMessage(message);

    if (CimExtendedTrace.cvTrcLvl) { cvTrace.internalCall(null,false,CLASS,METHOD); }
  }

}
